/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.asodpop.tests;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;

import ch.epfl.lia.frodo.algorithms.AgentFactory;
import ch.epfl.lia.frodo.algorithms.AgentInterface;
import ch.epfl.lia.frodo.algorithms.Problem;
import ch.epfl.lia.frodo.algorithms.XCSPparser;
import ch.epfl.lia.frodo.algorithms.dpop.DPOPsolver;
import ch.epfl.lia.frodo.algorithms.reformulation.ProblemRescaler;
import ch.epfl.lia.frodo.algorithms.test.AllTests;
import ch.epfl.lia.frodo.algorithms.asodpop.ASODPOP;
import ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface;
import ch.epfl.lia.frodo.communication.Message;
import ch.epfl.lia.frodo.communication.MessageWith2Payloads;
import ch.epfl.lia.frodo.communication.Queue;
import ch.epfl.lia.frodo.communication.QueueOutputPipeInterface;
import ch.epfl.lia.frodo.communication.mailer.CentralMailer;
import ch.epfl.lia.frodo.communication.sharedMemory.QueueIOPipe;
import ch.epfl.lia.frodo.controller.Controller;
import ch.epfl.lia.frodo.controller.WhitePages;
import ch.epfl.lia.frodo.daemon.LocalAgentReport;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import ch.epfl.lia.frodo.solutionSpaces.DCOPProblemInterface;
import junit.extensions.RepeatedTest;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * JUnit test of the ASODPOP agent
 * @author Brammert Ottens, Thomas Leaute
 * @param <V> the type used for variable values
 * @param <U> the type used for utility values
 *
 */
public class ASODPOPagentTest < V extends Addable<V>, U extends Addable<U> > extends TestCase implements IncomingMsgPolicyInterface<String> {


	/** Maximum number of variables in the problem 
	 * @note Must be at least 2. 
	 */
	private final int maxNbrVars = 10;
	
	/** Maximum number of binary constraints in the problem */
	private final int maxNbrEdges = 15;

	/** Maximum number of agents */
	private final int maxNbrAgents = 10;
	
	/** The queue used to listen to the agents */
	private Queue queue;
	
	/** For each agent, the output pipe to it */
	private Map<Object, QueueOutputPipeInterface> pipes;
	
	/** The testers pipe */
	QueueIOPipe pipe;
	
	/** All agents, indexed by their IDs */
	private Map< String, AgentInterface<V> > agents;
	
	/** Total number of agents */
	private int nbrAgents;
	
	/** Used to track the number of various types of messages received from the agents */
	private int nbrMsgsReceived;
	
	/** Number of agents finished */
	private int nbrAgentsFinished;
	
	/** Used to synchronize access to \a nbrAgentsFinished */
	private Object nbrAgentsFinished_lock = new Object ();

	/** The ADOPT stats gatherer listening for the solution */
	private ASODPOP<V, U> statsGatherer;
	
	/** The description of the agent */
	private Document agentDesc;
	
	/** \c true if the algorithm must be tested with the central mailer */
	private boolean useCentralMailer;
	
	/** \c true when the central mailer is to be tested with delays */
	private boolean useDelay;

	/** The CentralMailer */
	private CentralMailer mailman;
	
	/** The type used for variable values */
	private Class<V> domClass;
	
	/** The type used for utility values */
	private Class<U> utilClass;
	
	/** Maximize of minimize */
	private final boolean maximize;
	
	/** The sign of costs/utilities */
	private final int sign;

	/** Whether to use XCSP */
	private boolean useXCSP;
	
	/** Creates a JUnit test case corresponding to the input method
	 * @param string 			name of the method
	 * @param useXCSP 			whether to use XCSP
	 * @param useCentralMailer 	\c true when the central mailer should be used and tested
	 * @param useDelay 			\c true when the algorithm should be tested with the use of delay
	 * @param domClass 			The type used for variable values
	 * @param utilClass 		the type used for utility values
	 * @param maximize 			Maximize of minimize
	 * @param sign 				The sign of costs/utilities
	 */
	public ASODPOPagentTest(String string, boolean useXCSP, boolean useCentralMailer, boolean useDelay, Class<V> domClass, Class<U> utilClass, boolean maximize, int sign) {
		super (string);
		this.useXCSP = useXCSP;
		this.useCentralMailer = useCentralMailer;
		this.useDelay = useDelay;
		this.domClass = domClass;
		this.utilClass = utilClass;
		this.maximize = maximize;
		this.sign = sign;
	}

	/** @return the test suite */
	public static TestSuite suite () {
		TestSuite suite = new TestSuite ("Random tests for ASODPOPagent");
		
		TestSuite tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", true, false, false, AddableInteger.class, AddableInteger.class, true, +1), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities and without XCSP");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", false, false, false, AddableInteger.class, AddableInteger.class, true, +1), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities on maximization problems with unrestricted utility signs");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", true, false, false, AddableInteger.class, AddableInteger.class, true, 0), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities on maximization problems with unrestricted utility signs and without XCSP");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", false, false, false, AddableInteger.class, AddableInteger.class, true, 0), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities on minimization problems with unrestricted cost signs");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", true, false, false, AddableInteger.class, AddableInteger.class, false, 0), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities on minimization problems with unrestricted cost signs and without XCSP");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", false, false, false, AddableInteger.class, AddableInteger.class, false, 0), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with integer utilities and real-valued variables");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableReal, AddableInteger> ("testRandomSharedMemory", true, false, false, AddableReal.class, AddableInteger.class, true, +1), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with real utilities");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableReal> ("testRandomSharedMemory", true, false, false, AddableInteger.class, AddableReal.class, true, +1), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes with real utilities and without XCSP");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableReal> ("testRandomSharedMemory", false, false, false, AddableInteger.class, AddableReal.class, true, +1), 500));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using QueueIOPipes and the central mailer with integer utilities");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", true, true, false, AddableInteger.class, AddableInteger.class, true, +1), 500));
		suite.addTest(tmp);
		
		/// @bug Rarely gets stuck
//		tmp = new TestSuite ("Tests using QueueIOPipes and the central mailer with delays with integer utilities");
//		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomSharedMemory", true, true, true, AddableInteger.class, AddableInteger.class, true, +1), 5000));
//		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests using TCP pipes with integer utilities");
		tmp.addTest(new RepeatedTest (new ASODPOPagentTest<AddableInteger, AddableInteger> ("testRandomTCP", true, false, false, AddableInteger.class, AddableInteger.class, true, +1), 500));
		suite.addTest(tmp);
		
		return suite;
	}
	
	/** @see junit.framework.TestCase#setUp() */
	@SuppressWarnings("unchecked")
	public void setUp () throws Exception {
		nbrMsgsReceived = 0;
		nbrAgentsFinished = 0;
		
		// Create the queue
		if (this.useCentralMailer) {
			mailman = new CentralMailer (false, this.useDelay, null);
			this.queue = mailman.newQueue(AgentInterface.STATS_MONITOR);
		} else 
			queue = new Queue (false);
		
		queue.addIncomingMessagePolicy(this);
		pipes = new HashMap<Object, QueueOutputPipeInterface> ();
		
		agentDesc = XCSPparser.parse("src/ch/epfl/lia/frodo/algorithms/asodpop/ASODPOPagent.xml", false);
		
		// Fix the ProblemRescaler's shift if the problem is not a minimization problem with non-negative costs
		if (this.maximize || this.sign <= 0) {
			for (Element module : (List<Element>) this.agentDesc.getRootElement().getChild("modules").getChildren()) {
				if (module.getAttributeValue("className").equals(ProblemRescaler.class.getName())) {
					module.setAttribute("shift", "1000000");
					break;
				}
			}
		}
	}
	
	/** @see junit.framework.TestCase#tearDown() */
	protected void tearDown () throws Exception {
		super.tearDown();
		if (this.useCentralMailer) 
			mailman.end();
		queue.end();
		for (QueueOutputPipeInterface pipe : pipes.values())
			pipe.close();
		pipes.clear();
		for (AgentInterface<V> agent : agents.values()) 
			agent.kill();
		agents.clear();
		queue = null;
		pipes = null;
		agents = null;
		statsGatherer = null;
		this.agentDesc = null;
		this.pipe = null;
	}
	
	/** Tests the ASODPOPagent on a random problem using QueueIOPipes for inter-agent communication
	 * @throws Exception 	if an error occurs
	 */
	@SuppressWarnings("unchecked")
	public void testRandomSharedMemory () throws Exception {
		testRandom(false, (Class<? extends XCSPparser<V, U>>) XCSPparser.class, false);
	}

	/** Tests the ASODPOPagent on a random problem using TCP pipes for inter-agent communication
	 * @throws Exception 	if an error occurs
	 */
	@SuppressWarnings("unchecked")
	public void testRandomTCP () throws Exception {
		testRandom(true, (Class<? extends XCSPparser<V, U>>) XCSPparser.class, false);
	}
	
	/** Tests the DPOPagent on a random problem 
	 * @param useTCP 						whether TCP pipes should be used for communication between agents
	 * @param parserClass					the parser to be used
	 * @param aggregate						\c true when we use a model with aggregate constraints
	 * @throws IOException 					if unable to access DPOPagent.xml
	 * @throws JDOMException 				if unable to parse DPOPagent.xml
	 * @throws NoSuchMethodException 		if the parser does not have a constructor that takes in a Document
	 * @throws InvocationTargetException 	if the parser constructor throws an exception
	 * @throws IllegalAccessException 		if the parser constructor is not accessible
	 * @throws InstantiationException 		if the parser is abstract
	 * @throws IllegalArgumentException 	should never happen
	 */
	public void testRandom (boolean useTCP, Class< ? extends XCSPparser<V, U> > parserClass, boolean aggregate) 
	throws JDOMException, IOException, NoSuchMethodException, IllegalArgumentException, 
	InstantiationException, IllegalAccessException, InvocationTargetException {
		// Set up the input pipe for the queue
		pipe = new QueueIOPipe (queue);
		
		// Create the problem and agent descriptions
		Document problemDoc = AllTests.createRandProblem(maxNbrVars, maxNbrEdges, maxNbrAgents, this.maximize, this.sign);
		
		// Instantiate the parser/subsolver
		Constructor< ? extends XCSPparser<V, U> > constructor = null;
		DCOPProblemInterface<V, U> problem = null;
		constructor = parserClass.getConstructor(Document.class);
		problem = constructor.newInstance(problemDoc);
		problem.setDomClass(domClass);
		problem.setUtilClass(utilClass);
		if (!this.useXCSP) {
			Problem<V, U> prob = new Problem<V, U> (this.maximize);
			prob.reset(problem);
			prob.setDomClass(domClass);
			prob.setUtilClass(utilClass);
			problem = prob;
		}
		
		statsGatherer = new ASODPOP<V, U> (null, problem);
		statsGatherer.setSilent(true);
		statsGatherer.getStatsFromQueue(queue);

		String useCentralMailerString = Boolean.toString(useCentralMailer);
		agentDesc.getRootElement().setAttribute("measureTime", useCentralMailerString);
		if(agentDesc.getRootElement().getChild("parser") != null)
			agentDesc.getRootElement().getChild("parser").setAttribute("parserClass", parserClass.getCanonicalName());
		else {
			Element elmt = new Element("parser");
			elmt.setAttribute("parserClass", parserClass.getCanonicalName());
			agentDesc.getRootElement().addContent(elmt);
		}
		
		// Go through the list of agents and instantiate them
		Set<String> agentNames = problem.getAgents();
		nbrAgents = agentNames.size();
		agents = new HashMap< String, AgentInterface<V> > (nbrAgents);
		synchronized (agents) {
			if (useTCP) { // use TCP pipes
				int port = 5500;
				for (String agent : agentNames) {
					DCOPProblemInterface<V, U> agentElmt = problem.getSubProblem(agent);
					agents.put(agentElmt.getAgent(), AgentFactory.createAgent(pipe, pipe, agentElmt, agentDesc, port++));
				}
			} else { // use QueueIOPipes
				for (String agent : agentNames) {
					DCOPProblemInterface<V, U> agentElmt = problem.getSubProblem(agent);
					agents.put(agentElmt.getAgent(), AgentFactory.createAgent(pipe, agentElmt, agentDesc, mailman));
				}
			}
		}
		
		if (this.useCentralMailer) 
			mailman.start();
		
		// Wait for all agents to finish
		long start = System.currentTimeMillis();
		while (true) {
			synchronized (nbrAgentsFinished_lock) {
				if (nbrAgentsFinished >= nbrAgents) {
					break;
				} else if (System.currentTimeMillis() - start > 10000) {
					fail("Timeout");
				}
			}
		}
		
		// Check that ASODPOP and DPOP agree on the total optimal utility
		U totalOptUtil = statsGatherer.getTotalOptUtil();
		Document dpopAgent = XCSPparser.parse(AgentFactory.class.getResourceAsStream("/ch/epfl/lia/frodo/algorithms/dpop/DPOPagent.xml"), false);
		dpopAgent.getRootElement().getChild("parser").setAttribute("utilClass", this.utilClass.getName());
		U dpopOptUtil = (new DPOPsolver<V, U> (dpopAgent, this.domClass, this.utilClass)).solve(problemDoc).getUtility();
		assertEquals (dpopOptUtil, totalOptUtil);
		
		// Check that the optimal assignments have indeed the reported utility
		Map<String, V> optAssignments = statsGatherer.getOptAssignments();
		assertEquals (totalOptUtil, problem.getUtility(optAssignments).getUtility(0));
	}

	/** @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#getMsgTypes() */
	public Collection<String> getMsgTypes() {
		ArrayList<String> types = new ArrayList<String> (4);
		types.add(AgentInterface.LOCAL_AGENT_REPORTING);
		types.add(AgentInterface.LOCAL_AGENT_ADDRESS_REQUEST);
		types.add(AgentInterface.AGENT_CONNECTED);
		types.add(AgentInterface.AGENT_FINISHED);
		return types;
	}

	/** @see IncomingMsgPolicyInterface#notifyIn(Message) */
	@SuppressWarnings("unchecked")
	public void notifyIn(Message msg) {
		
		if (msg.getType().equals(AgentInterface.LOCAL_AGENT_REPORTING)) {
			LocalAgentReport msgCast = (LocalAgentReport) msg;
			String agentID = msgCast.getAgentID();
			QueueIOPipe pipe = msgCast.getLocalPipe();
			queue.addOutputPipe(agentID, pipe);
			
			// Create a TCP pipe if required
			int port = msgCast.getPort();
			if (port >= 0) {
				try {
					pipes.put(agentID, Controller.PipeFactoryInstance.outputPipe(Controller.PipeFactoryInstance.getSelfAddress(port)));
				} catch (UnknownHostException e) { // should never happen
					e.printStackTrace();
					fail();
				} catch (IOException e) {
					e.printStackTrace();
					fail();
				}
			} else 
				pipes.put(agentID, pipe);
			
			// Check if all agents have reported, and if so, tell them to connect
			if (pipes.size() >= nbrAgents) {
				synchronized (agents) { // synchronize so we don't tell the agents to connect before the list of agents is ready
					queue.sendMessageToMulti(pipes.keySet(), new Message (WhitePages.CONNECT_AGENT));
				}
			}
		}
		
		else if (msg.getType().equals(AgentInterface.LOCAL_AGENT_ADDRESS_REQUEST)) {
			MessageWith2Payloads<String, String> msgCast = (MessageWith2Payloads<String, String>) msg;
			String recipient = msgCast.getPayload2();
			agents.get(msgCast.getPayload1()).addOutputPipe(recipient, pipes.get(recipient));
		}
		
		else if (msg.getType().equals(AgentInterface.AGENT_CONNECTED)) {
			if (++nbrMsgsReceived >= nbrAgents) { // all agents are now connected; tell them to start
				queue.sendMessageToMulti(pipes.keySet(), new Message (AgentInterface.START_AGENT));
			}
		}
		
		else if (msg.getType().equals(AgentInterface.AGENT_FINISHED)) {
			synchronized (nbrAgentsFinished_lock) {
				nbrAgentsFinished++;
				if (this.useCentralMailer) // the granularity of the timer is such that the > test sometimes fails
					assertTrue (queue.getCurrentMessageWrapper().getTime() >= 0);
			}
		}
		
	}

	/** Does nothing
	 * @see ch.epfl.lia.frodo.communication.IncomingMsgPolicyInterface#setQueue(ch.epfl.lia.frodo.communication.Queue)
	 */
	public void setQueue(Queue queue) { }

}
