/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

package ch.epfl.lia.frodo.algorithms.subsolvers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import junit.extensions.RepeatedTest;
import junit.framework.TestSuite;
import ch.epfl.lia.frodo.algorithms.RandGraphFactory;
import ch.epfl.lia.frodo.algorithms.XCSPparser;
import ch.epfl.lia.frodo.algorithms.subsolvers.SimpleSubsolver.CompoundSpace;
import ch.epfl.lia.frodo.algorithms.test.AllTests;
import ch.epfl.lia.frodo.algorithms.test.XCSPparserTest;
import ch.epfl.lia.frodo.solutionSpaces.AddableInteger;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import ch.epfl.lia.frodo.solutionSpaces.DCOPProblemInterface;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.Hypercube;
import ch.epfl.lia.frodo.solutionSpaces.hypercube.ScalarHypercube;

/** Test suite for SimpleSubsolver
 * @author Thomas Leaute
 */
public class SimpleSubsolverTest extends XCSPparserTest {

	/** Maximum number of variables in the random graph 
	 * @note Must be at least 2. 
	 */
	private final int maxNbrVars = 10;
	
	/** Maximum number of edges in the random graph */
	private final int maxNbrEdges = 50;

	/** Maximum number of agents */
	private final int maxNbrAgents = 5;

	/** Generates a test using the specified method
	 * @param name 	name of the method
	 */
	public SimpleSubsolverTest(String name) {
		super(name);
	}

	/** @see XCSPparserTest#setUp() */
	protected void setUp() {
		graph = RandGraphFactory.getRandGraph(maxNbrVars, maxNbrEdges, maxNbrAgents);
		solutionSpaces = new ArrayList< Hypercube<AddableInteger, AddableInteger> > ();
		probSpaces = new ArrayList< Hypercube<AddableInteger, AddableReal> > ();
		problem = AllTests.generateProblem(graph, graph.nodes.size(), solutionSpaces, probSpaces, true);
		parserInt = new SimpleSubsolver<AddableInteger, AddableInteger> (problem);
		parserInt.setUtilClass(AddableInteger.class);
		parserReal = new SimpleSubsolver<AddableInteger, AddableReal> (problem);
		parserReal.setUtilClass(AddableReal.class);
	}

	/** @return the test suite for this test */
	public static TestSuite suite () {
		
		TestSuite suite = new TestSuite ("JUnit tests for SimpleSubsolver");
		
		suite.addTest(new SimpleSubsolverTest ("testGetZeroUtility"));

		suite.addTest(new SimpleSubsolverTest ("testGetPlusInfUtility"));

		suite.addTest(new SimpleSubsolverTest ("testGetMinInfUtility"));
		
		TestSuite tmp = new TestSuite ("Tests for getAgents");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetAgents"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getOwner");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetOwner"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getOwners");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetOwners"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNbrVars");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNbrVars"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNbrVars that takes in the name of an agent");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNbrVarsString"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getVariables");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetVariables"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getVariables for an input agent");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetVariablesString"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getExtVars");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetExtVars"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNeighborVars");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNeighborVars"), 5000));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNbrNeighbors");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNbrNeighbors"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNeighborhoods");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNeighborhoods"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getNeighborhoodSizes");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetNeighborhoodSizes"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getAgentNeighborhoods");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetAgentNeighborhoods"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getDomainSize");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetDomainSize"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getDomain");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetDomain"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for setDomain");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testSetDomain"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getSubProblem");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetSubProblem"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for isRandom");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testIsRandom"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getSolutionSpaces with random variables");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetSolutionSpacesWithRandVars"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getSolutionSpaces ignoring random variables");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetSolutionSpaces"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getProbabilitySpaces");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetProbabilitySpaces"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getSolutionSpaces for a specific variable with random variables");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetSolutionSpacesForVarWithRandVars"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getSolutionSpaces for a specific variable ignoring random variables");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetSolutionSpacesForVar"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getProbabilitySpaces for a specific variable");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetProbabilitySpacesForVar"), 100));
		suite.addTest(tmp);
		
		tmp = new TestSuite ("Tests for getUtility on complete assignments");
		tmp.addTest(new RepeatedTest (new SimpleSubsolverTest ("testGetUtility"), 500));
		suite.addTest(tmp);

		return suite;
	}
	
	/** @see XCSPparserTest#testGetOwners() */
	@Override
	public void testGetOwners() {
		
		// First test on the overall problem
		Map<String, String> owners = parserInt.getOwners();
		ext: for (String var : graph.nodes) {
			
			// Check whether var is a backyard variable
			Integer owner = graph.clusterOf.get(var);
			for (String neigh : graph.neighborhoods.get(var)) {
				if (! owner.equals(graph.clusterOf.get(neigh))) { // not a backyard variable
					assertEquals (owner.toString(), owners.remove(var));
					continue ext;
				}
			}
			
			// var is a backyard variable, unless its component contains only backyard variables:
			// then exactly one of the variables must be in owners
			List<String> component = graph.components.get(graph.componentOf.get(var));
			for (String compNeigh : component) {
				if (! owner.equals(graph.clusterOf.get(compNeigh))) { // not a pure-backyard component
					assertTrue (! owners.containsKey(var));
					continue ext;
				}
			}
			
			// This variable belongs to a pure-backyard component
			// Check whether it is the first in lexicographic order
			for (String compNeigh : component) {
				if (var.compareTo(compNeigh) > 0) { // not first
					assertTrue (! owners.containsKey(var));
					continue ext;
				}
			}
			
			// First variable
			assertEquals (owner.toString(), owners.remove(var));
		}
		assertTrue (owners.isEmpty());
		
		// Now test on each agent's subproblem
		for (Integer i = 0; i < graph.clusters.size(); i++) {
			String agent = i.toString();
			
			DCOPProblemInterface<AddableInteger, AddableInteger> subProb = parserInt.getSubProblem(agent);
			owners = subProb.getOwners();
			
			// Go through all variables owned by the current agent, and examine their neighbors
			ext: for (String var : graph.clusters.get(i)) {
				
				// Check if the current variable is connected to at least one external variable
				for (String neigh : graph.neighborhoods.get(var)) {
					if (! i.equals(graph.clusterOf.get(neigh))) {
						
						assertEquals (agent, owners.remove(var));
						
						// Go through all external neighbors of the current variable
						for (String neigh2 : graph.neighborhoods.get(var)) {
							
							// Check the owner of this neighbor
							Integer owner = graph.clusterOf.get(neigh2);
							if (! i.equals(owner)) // this neighbor is an external variable
								assertEquals (owner.toString(), owners.get(neigh2));
						}

						// Examine the next variable
						continue ext;
					}
				}
				
				// var is a backyard variable, unless its component contains only backyard variables:
				// then exactly one of the variables must be in owners
				List<String> component = graph.components.get(graph.componentOf.get(var));
				for (String compNeigh : component) {
					if (! i.equals(graph.clusterOf.get(compNeigh))) { // not a pure-backyard component
						assertTrue (! owners.containsKey(var));
						continue ext;
					}
				}
				
				// This variable belongs to a pure-backyard component
				// Check whether it is the first in lexicographic order
				for (String compNeigh : component) {
					if (var.compareTo(compNeigh) > 0) { // not first
						assertTrue (! owners.containsKey(var));
						continue ext;
					}
				}
				
				// First variable
				assertEquals (i.toString(), owners.remove(var));
			}
			
			// All remaining variables must be external variables visible by the agent
			for (Map.Entry<String, String> entry : owners.entrySet()) {
				assertTrue (! entry.getValue().equals(agent));
				assertTrue (parserInt.getAgentNeighborhoods().get(entry.getKey()).contains(agent));
			}
		}
	}
	
	/** @return for each agent, the number of backyard variables it controls that need to be promoted to the facade */
	private int[] getNbrPromoted () {
		
		int[] promotedPerAgent = new int [graph.clusters.size()];
		ext: for (List<String> component : graph.components) {
			Iterator<String> iter = component.iterator();
			Integer owner = graph.clusterOf.get(iter.next());
			while (iter.hasNext()) 
				if (! owner.equals(graph.clusterOf.get(iter.next()))) // not a pure-backyard component
					continue ext;
			
			// Pure-backyard component
			promotedPerAgent[owner]++;
		}

		return promotedPerAgent;
	}

	/** @see XCSPparserTest#testGetNbrVars() */
	@Override
	public void testGetNbrVars() {
		
		// Count the number of pure-backyard components
		int[] promotedPerAgent = getNbrPromoted();
		int totalPromoted = 0;
		for (int i = 0; i < promotedPerAgent.length; i++) 
			totalPromoted += promotedPerAgent[i];
		
		// Test for the overall problem
		int nbrVars = 0;
		ext: for (String var : graph.nodes) {
			
			// Check whether var has an external neighbor
			Integer owner = graph.clusterOf.get(var);
			for (String neigh : graph.neighborhoods.get(var)) {
				if (! owner.equals(graph.clusterOf.get(neigh))) { // not a backyard variable
					nbrVars++;
					continue ext;
				}
			}
		}
		assertEquals (nbrVars + totalPromoted, parserInt.getNbrVars());
		
		// Test for each agent's subproblem
		for (Integer i = 0; i < graph.clusters.size(); i++) {
			
			// Compute the correct number of variables in this agent's subproblem
			HashSet<String> allVars = new HashSet<String> ();
			List<String> intVars = graph.clusters.get(i);
			ext: for (String intVar : intVars) {
				
				// Check that this variable is not a backyard variable
				for (String neigh : graph.neighborhoods.get(intVar)) {
					if (! i.equals(graph.clusterOf.get(neigh))) {
						
						allVars.add(intVar);
						
						// Go through the list of external neighbors of the current variable
						for (String neigh2 : graph.neighborhoods.get(intVar)) {
							
							// Check the owner of this neighbor
							Integer owner = graph.clusterOf.get(neigh2);
							if (! i.equals(owner)) // external variable
								allVars.add(neigh2);
						}
						
						continue ext;
					}
				}
				
			}
			
			assertEquals (allVars.size() + promotedPerAgent[i], parserInt.getSubProblem(i.toString()).getNbrVars());
		}
	}

	/** @see XCSPparserTest#testGetNbrVarsString() */
	@Override
	public void testGetNbrVarsString() {
		
		// Compute each agent's number of facade variables
		int[] promotedPerAgent = getNbrPromoted();
		int[] facades = promotedPerAgent.clone();
		for (Integer i = 0; i < graph.clusters.size(); i++) {

			// Go through all internal variables, looking for backyard variables
			ext: for (String var : graph.clusters.get(i)) {
				
				// Go through all neighboring variables until we find one external variable
				for (String neigh : graph.neighborhoods.get(var)) {
					if (! i.equals(graph.clusterOf.get(neigh))) {
						facades[i]++;
						continue ext;
					}
				}
			}
		}
		
		// Check for the overall problem
		for (int i = 0; i < graph.clusters.size(); i++) 
			assertEquals (facades[i], parserInt.getNbrVars(String.valueOf(i)));
		
		// Check for each agent's subproblem
		for (Integer i = 0; i < graph.clusters.size(); i++) {
			
			SimpleSubsolver<AddableInteger, AddableInteger> subprob = (SimpleSubsolver<AddableInteger, AddableInteger>) parserInt.getSubProblem(i.toString());
			
			// Check for each input agent
			for (Integer j = 0; j < graph.clusters.size(); j++) {
				
				// Check if the two agents are the same
				if (i.equals(j)) {
					
					// Count the number of facade variables
					int nbrVars = 0;
					ext: for (String var : graph.clusters.get(i)) { // for each internal variable
						
						// Check that var has at least one external neighbor
						for (String neigh : graph.neighborhoods.get(var)) {
							if (! i.equals(graph.clusterOf.get(neigh))) { // external neighbor
								nbrVars++;
								continue ext;
							}
						}
					}
					
					assertEquals (nbrVars + promotedPerAgent[i],subprob.getNbrVars(j.toString()));
				}
				else { // the two agents are different
					
					// Count the number of variables owned by j that are visible by i
					int nbrVars = 0;
					ext: for (String var : graph.clusters.get(j)) {
						for (String neigh : graph.neighborhoods.get(var)) {
							if (i.equals(graph.clusterOf.get(neigh))) { // neigh is owned by i
								nbrVars++;
								continue ext;
							}
						}
					}
					
					assertEquals (nbrVars, subprob.getNbrVars(j.toString()));
				}
			}

		}
	}

	/** @see XCSPparserTest#testGetVariables() */
	@Override
	public void testGetVariables() {
		
		// Test for the overall problem
		Collection<String> vars = parserInt.getVariables();
		ext: for (String var : graph.nodes) {
			
			// Check whether var has an external neighbor
			Integer owner = graph.clusterOf.get(var);
			for (String neigh : graph.neighborhoods.get(var)) {
				if (! owner.equals(graph.clusterOf.get(neigh))) { // not a backyard variable
					assertTrue(vars.remove(var));
					continue ext;
				}
			}
		}
		
		// Each remaining variable must be the only (promoted to) facade variable in a pure-backyard component
		// and it must be the first in lexicographic order
		for (String var : vars) {
			Integer owner = graph.clusterOf.get(var);
			for (String compNeigh : graph.components.get(graph.componentOf.get(var))) {
				if (! compNeigh.equals(var)) {
					assertEquals (owner, graph.clusterOf.get(compNeigh));
					assertTrue (! vars.contains(compNeigh));
					assertTrue (var.compareTo(compNeigh) < 0);
				}
			}
		}
		
		// Test for each agent's subproblem
		for (Integer i = 0; i < graph.clusters.size(); i++) {
			
			vars = parserInt.getSubProblem(String.valueOf(i)).getVariables();
			
			// Compute the correct number of variables in this agent's subproblem
			HashSet<String> allVars = new HashSet<String> ();
			List<String> intVars = graph.clusters.get(i);
			ext: for (String intVar : intVars) {
				
				// Check that this variable is not a backyard variable
				for (String neigh : graph.neighborhoods.get(intVar)) {
					if (! i.equals(graph.clusterOf.get(neigh))) {
						
						allVars.add(intVar);
						
						// Go through the list of external neighbors of the current variable
						for (String neigh2 : graph.neighborhoods.get(intVar)) {
							
							// Check the owner of this neighbor
							Integer owner = graph.clusterOf.get(neigh2);
							if (! i.equals(owner)) // external variable
								allVars.add(neigh2);
						}
						
						continue ext;
					}
				}
			}
			
			for (String var : allVars) 
				assertTrue (vars.remove(var));

			// Each remaining variable must be the only (promoted to) facade variable in a pure-backyard component
			// and it must be the first in lexicographic order
			for (String var : vars) {
				Integer owner = graph.clusterOf.get(var);
				assertEquals (i, owner);
				for (String compNeigh : graph.components.get(graph.componentOf.get(var))) {
					if (! compNeigh.equals(var)) {
						assertEquals (owner, graph.clusterOf.get(compNeigh));
						assertTrue (! vars.contains(compNeigh));
						assertTrue (var.compareTo(compNeigh) < 0);
					}
				}
			}
		}
	}

	/** @see XCSPparserTest#testGetVariablesString() */
	@Override
	public void testGetVariablesString() {
		
		// Test for each agent
		for (Integer i = 0; i < graph.clusters.size(); i++) {
			
			Collection<String> vars = parserInt.getVariables(i.toString());
			
			// Look for facade variables only
			ext: for (String var : graph.clusters.get(i)) {
				
				// Check if var has at least one external neighbor
				for (String neigh : graph.neighborhoods.get(var)) {
					if (! i.equals(graph.clusterOf.get(neigh))) {
						assertTrue (vars.remove(var));
						continue ext;
					}
				}
			}
			
			// Each remaining variable must be the only (promoted to) facade variable in a pure-backyard component
			// and it must be the first in lexicographic order
			for (String var : vars) {
				Integer owner = graph.clusterOf.get(var);
				assertEquals (i, owner);
				for (String compNeigh : graph.components.get(graph.componentOf.get(var))) {
					if (! compNeigh.equals(var)) {
						assertEquals (owner, graph.clusterOf.get(compNeigh));
						assertTrue (! vars.contains(compNeigh));
						assertTrue (var.compareTo(compNeigh) < 0);
					}
				}
			}
		}
		
		// Test for anonymous variables
		Collection<String> vars = parserInt.getVariables(null);
		for (Hypercube<AddableInteger, AddableReal> probSpace : this.probSpaces) 
			assertTrue (vars.remove(probSpace.getVariable(0)));
		assertTrue (vars.isEmpty());
	}
	
	/** @see XCSPparserTest#getNeighborhoods(boolean) */
	@Override
	protected Map< String, ? extends Collection<String> > getNeighborhoods(boolean withAnonymVars) {

		HashMap< String, HashSet<String> > out = new HashMap< String, HashSet<String> > ();
		
		// Loop while there are remaining variables
		HashSet<String> allVars = new HashSet<String> (graph.nodes);
		ext: while (! allVars.isEmpty()) {
			
			// Pop the first variable
			Iterator<String> iter = allVars.iterator();
			String var = iter.next();
			iter.remove();
			
			// Consider this variable if it is a facade variable
			Integer owner = graph.clusterOf.get(var);
			for (String neigh : graph.neighborhoods.get(var)) {
				if (! owner.equals(graph.clusterOf.get(neigh))) { // var has an external neighbor 
					
					// Recursively traverse the graph, looking for facade neighbors of var
					HashSet<String> neighbors = new HashSet<String> ();
					this.getFacadeNeighborhoods(var, neighbors, new HashSet<String> ());
					neighbors.remove(var);
					
					// Record all facade neighbors of var
					out.put(var, new HashSet<String> (neighbors));
					
					continue ext;
				}
			}
			
			// This variable is in the backyard; compute its component
			HashSet<String> facadeComp = new HashSet<String> ();
			HashSet<String> backyardComp = new HashSet<String> ();
			this.getFacadeNeighborhoods(var, facadeComp, backyardComp);
			
			// Check if this variable is in a pure-backyard component
			if (facadeComp.isEmpty()) {
				
				// Find the lexicographically first variable 
				String first = var;
				for (String var2 : backyardComp) 
					if (first.compareTo(var2) > 0) 
						first = var2;
				
				// Promote the first variable, and prune out the others
				out.put(first, new HashSet<String> ());
				allVars.removeAll(backyardComp);
			}
		}
		
		// No more variables to explore
		return out;
	}
	
	/** Computes the facade neighborhoods in the problem
	 * @param var 			the current variable 
	 * @param neighbors 	the current set of facade neighbors
	 * @param backyard 		the current set of backyard variables 
	 */
	private void getFacadeNeighborhoods (String var, HashSet<String> neighbors, HashSet<String> backyard) {
		
		// Look at all neighbors of var
		ext: for (String neigh : graph.neighborhoods.get(var)) {
			
			// Check if neigh is a facade variable
			Integer owner = graph.clusterOf.get(neigh);
			for (String neigh2 : graph.neighborhoods.get(neigh)) {
				if (! owner.equals(graph.clusterOf.get(neigh2))) { // neigh has an external neighbor
					neighbors.add(neigh);
					continue ext;
				}
			}
			
			// neigh is a backyard variable
			if (backyard.add(neigh))  // neigh has not already been visited
				this.getFacadeNeighborhoods(neigh, neighbors, backyard);
		}
	}

	/** @see XCSPparserTest#testGetSolutionSpaces(XCSPparser, List, String, boolean) */
	@Override
	protected void testGetSolutionSpaces(XCSPparser<AddableInteger, AddableInteger> parser, 
			List< ? extends UtilitySolutionSpace<AddableInteger, AddableInteger> > refSpaces, String var, boolean withRandVars) {
		
		// Go through all compound solution spaces in the problem
		for (UtilitySolutionSpace<AddableInteger, AddableInteger> space : parser.getSolutionSpaces(var, withRandVars)) {
			
			// Test that the space officially does not involve any backyard variables
			Map< String, ? extends Collection<String> > neighborhoods = this.getNeighborhoods(withRandVars);
			HashSet<String> vars = new HashSet<String> ();
			for (String var2 : space.getVariables()) {
				vars.add(var2);
				assertTrue (parser.isRandom(var2) || neighborhoods.containsKey(var2));
			}

			if (! (space instanceof CompoundSpace<?, ?>)) { // the space is a hypercube
				assertTrue (refSpaces.remove(space));
			}
			else { // the space is a compound space
				
				CompoundSpace<AddableInteger, AddableInteger> spaceCast = (CompoundSpace<AddableInteger, AddableInteger>) space;
				
				// Go through the list of spaces in this compound space
				String agent = null;
				for (UtilitySolutionSpace<AddableInteger, AddableInteger> subspace : spaceCast.getSpaces()) {
					
					assertTrue (refSpaces.remove(subspace));
					
					// Go through the list of variables in this subspace
					for (String var2 : subspace.getVariables()) {
						
						// The variable must be either officially in the compound space, or a backyard variable
						if (! vars.contains(var2)) {
							if (agent == null) {
								agent = parser.getOwner(var2);
							} else {
								assertEquals (agent, parser.getOwner(var2));
								
								// The variable must not be connected to any external variable
								for (String neigh : graph.neighborhoods.get(var2)) 
									assertEquals (agent, parser.getOwner(neigh));
							}
						}
					}
				}
			}
		}
		
		// Check that the only spaces remaining in refSpaces do not contain the variable
		if (var != null) 
			for (UtilitySolutionSpace<AddableInteger, AddableInteger> space : refSpaces) 
				assertNull (space.getDomain(var));
	}

	/** @see ch.epfl.lia.frodo.algorithms.test.XCSPparserTest#testGetUtility() */
	@Override
	public void testGetUtility () {
		
		// Choose a random assignment to all variables
		int nbrVars = parserInt.getNbrVars();
		String[] vars = parserInt.getVariables().toArray(new String [nbrVars]);
		AddableInteger[] vals = new AddableInteger [nbrVars];
		HashMap<String, AddableInteger> assignments = new HashMap<String, AddableInteger> ();
		for (int i = 0; i < nbrVars; i++) {
			String var = vars[i];
			AddableInteger[] domain = parserInt.getDomain(var);
			AddableInteger val = domain[ (int) (domain.length * Math.random()) ];
			vals[i] = val;
			assignments.put(var, val);
		}
		
		AddableInteger util = parserInt.getUtility(assignments).getUtility(0);
		
		// Compute the true utility
		UtilitySolutionSpace<AddableInteger, AddableInteger>  trueUtil = new ScalarHypercube<AddableInteger, AddableInteger> (new AddableInteger (0), 
				(parserInt.maximize() ? parserInt.getMinInfUtility() : parserInt.getPlusInfUtility()), new AddableInteger [0].getClass());
		ext: for (UtilitySolutionSpace<AddableInteger, AddableInteger> space : this.solutionSpaces) {
			
			// Skip this space if it contains a random variable
			for (String var : space.getVariables()) 
				if (! graph.nodes.contains(var)) 
					continue ext;
			
			trueUtil = trueUtil.join(space.slice(vars, vals));
		}
		
		assertEquals (trueUtil.blindProjectAll(parserInt.maximize()), util); // project out all backyard variables
	}
	
}
